#include <std_include.hpp>
#include "loader/component_loader.hpp"

#include <utils/binary_resource.hpp>
#include <utils/nt.hpp>
#include <utils/string.hpp>

#include <resource.hpp>

#include "steam/steam.hpp"
#include "steam/interface.hpp"

#include "scheduler.hpp"

#include <xorstr.hpp>

namespace {
enum class ownership_state {
  success,
  unowned,
  nosteam,
  error,
};

ownership_state state_;

utils::binary_resource runner_file(RUNNER, xorstr_("iw4sp-runner.exe"));
} // namespace

class steam_proxy final : public component_interface {
public:
  void post_load() override {
    this->load_client();
    this->clean_up_on_error();

    try {
      state_ = this->start_mod(xorstr_("iw4x-sp singleplayer"), 10180);
    } catch (const std::exception& ex) {
      state_ = ownership_state::error;
      printf(xorstr_("Steam: %s\n"), ex.what());
    }

#ifdef STEAM_OWNERSHIP_CHECK
    if (utils::nt::is_wine()) {
      state_ = ownership_state::success;
      return;
    }
#endif

#ifdef STEAM_OWNERSHIP_CHECK
    evaluate_ownership_state(state_);
#endif
  }

  void pre_destroy() override {
    if (this->steam_client_module_) {
      if (this->steam_pipe_) {
        if (this->global_user_) {
          this->steam_client_module_.invoke<void>(xorstr_("Steam_ReleaseUser"),
                                                  this->steam_pipe_,
                                                  this->global_user_);
        }

        this->steam_client_module_.invoke<bool>(
            xorstr_("Steam_BReleaseSteamPipe"), this->steam_pipe_);
      }
    }
  }

private:
  utils::nt::library steam_client_module_{};

  steam::interface client_engine_{};
  steam::interface client_user_{};
  steam::interface client_utils_{};

  void* steam_pipe_ = nullptr;
  void* global_user_ = nullptr;

  [[nodiscard]] void* load_client_engine() const {
    if (!this->steam_client_module_)
      return nullptr;

    for (auto i = 1; i < 1000; ++i) {
      const auto* name =
          utils::string::va(xorstr_("CLIENTENGINE_INTERFACE_VERSION{0:03}"), i);

      auto* const client_engine = this->steam_client_module_.invoke<void*>(
          xorstr_("CreateInterface"), name, nullptr);
      if (client_engine)
        return client_engine;
    }

    return nullptr;
  }

  void load_client() {
    const auto steam_path = steam::get_steam_install_directory();
    if (steam_path.empty())
      return;

    utils::nt::library::load(steam_path / xorstr_("tier0_s.dll"));
    utils::nt::library::load(steam_path / xorstr_("vstdlib_s.dll"));
    this->steam_client_module_ =
        utils::nt::library::load(steam_path / xorstr_("steamclient.dll"));
    if (!this->steam_client_module_)
      return;

    this->client_engine_ = load_client_engine();
    if (!this->client_engine_)
      return;

    this->steam_pipe_ = this->steam_client_module_.invoke<void*>(
        xorstr_("Steam_CreateSteamPipe"));
    this->global_user_ = this->steam_client_module_.invoke<void*>(
        xorstr_("Steam_ConnectToGlobalUser"), this->steam_pipe_);
    this->client_user_ = this->client_engine_.invoke<void*>(
        8, this->steam_pipe_, this->global_user_); // GetIClientUser
    this->client_utils_ = this->client_engine_.invoke<void*>(
        14, this->steam_pipe_); // GetIClientUtils
  }

  ownership_state start_mod(const std::string& title,
                            const std::size_t app_id) {
    __try {
      return this->start_mod_unsafe(title, app_id);
    } __except (EXCEPTION_EXECUTE_HANDLER) {
      this->do_cleanup();
      return ownership_state::error;
    }
  }

  ownership_state start_mod_unsafe(const std::string& title,
                                   std::size_t app_id) {
    if (!this->client_utils_ || !this->client_user_)
      return ownership_state::nosteam;

    if (!this->client_user_.invoke<bool>(xorstr_("BIsSubscribedApp"), app_id)) {
#ifdef _DEBUG
      app_id = 480; // Spacewar
#else
      return ownership_state::unowned;
#endif
    }

    this->client_utils_.invoke<void>(xorstr_("SetAppIDForCurrentPipe"), app_id,
                                     false);

    char our_directory[MAX_PATH]{};
    GetCurrentDirectoryA(sizeof(our_directory), our_directory);

    const auto path = runner_file.get_extracted_file();
    const auto* cmd_line = utils::string::va(xorstr_("\"{0}\" -proc {1}"), path,
                                             GetCurrentProcessId());

    game_id game_id;
    game_id.raw.type = 1; // k_EGameIDTypeGameMod
    game_id.raw.app_id = app_id & 0xFFFFFF;

    const auto* mod_id = "IW4x-SP";
    game_id.raw.mod_id =
        *reinterpret_cast<const unsigned int*>(mod_id) | 0x80000000;

    this->client_user_.invoke<bool>(xorstr_("SpawnProcess"), path.c_str(),
                                    cmd_line, our_directory, game_id.bits,
                                    title.c_str(), app_id, 0, 0, 0);

    return ownership_state::success;
  }

  void do_cleanup() {
    this->client_engine_ = nullptr;
    this->client_user_ = nullptr;
    this->client_utils_ = nullptr;

    this->steam_pipe_ = nullptr;
    this->global_user_ = nullptr;

    this->steam_client_module_ = utils::nt::library{nullptr};
  }

  void clean_up_on_error() {
    scheduler::schedule(
        [this] {
          if (this->steam_client_module_ && this->steam_pipe_ &&
              this->global_user_ &&
              this->steam_client_module_.invoke<bool>(
                  xorstr_("Steam_BConnected"), this->global_user_,
                  this->steam_pipe_) &&
              this->steam_client_module_.invoke<bool>(
                  xorstr_("Steam_BLoggedOn"), this->global_user_,
                  this->steam_pipe_)) {
            return scheduler::cond_continue;
          }

          this->do_cleanup();
          return scheduler::cond_end;
        },
        scheduler::pipeline::async);
  }

#ifdef STEAM_OWNERSHIP_CHECK
  void evaluate_ownership_state(const ownership_state state) {
    switch (state) {
    case ownership_state::nosteam:
      throw std::runtime_error(
          xorstr_("Steam must be running to play this game!"));
    case ownership_state::unowned:
      throw std::runtime_error(
          xorstr_("You must own the game on Steam to play this mod!"));
    case ownership_state::error:
      throw std::runtime_error(
          xorstr_("Failed to verify ownership of the game!"));
    case ownership_state::success:
      break;
    }
  }
#endif
};

REGISTER_COMPONENT(steam_proxy)
